from textwrap import dedent

import pytest
from mcp import StdioServerParameters

from mcpadapt.core import MCPAdapt
from mcpadapt.langchain_adapter import LangChainAdapter


@pytest.fixture
def echo_server_script():
    return dedent(
        '''
        from mcp.server.fastmcp import FastMCP

        mcp = FastMCP("Echo Server")

        @mcp.tool()
        def echo_tool(text: str) -> str:
            """Echo the input text"""
            return f"Echo: {text}"
        
        mcp.run()
        '''
    )


@pytest.mark.asyncio
async def test_basic_async(echo_server_script):
    async with MCPAdapt(
        StdioServerParameters(
            command="uv", args=["run", "python", "-c", echo_server_script]
        ),
        LangChainAdapter(),
    ) as tools:
        assert len(tools) == 1  # we expect one tool as defined above
        assert tools[0].name == "echo_tool"  # we expect the tool to be named echo_tool
        response = await tools[0].ainvoke("hello")
        assert response == "Echo: hello"  # we expect the tool to return "Echo: hello"


def test_basic_sync(echo_server_script):
    with MCPAdapt(
        StdioServerParameters(
            command="uv", args=["run", "python", "-c", echo_server_script]
        ),
        LangChainAdapter(),
    ) as tools:
        assert len(tools) == 1
        assert tools[0].name == "echo_tool"
        assert tools[0].invoke("hello") == "Echo: hello"


def test_tool_name_with_dashes():
    mcp_server_script = dedent(
        '''
        from mcp.server.fastmcp import FastMCP

        mcp = FastMCP("Echo Server")

        @mcp.tool(name="echo-tool")
        def echo_tool(text: str) -> str:
            """Echo the input text"""
            return f"Echo: {text}"
        
        mcp.run()
        '''
    )
    with MCPAdapt(
        StdioServerParameters(
            command="uv", args=["run", "python", "-c", mcp_server_script]
        ),
        LangChainAdapter(),
    ) as tools:
        assert len(tools) == 1
        assert tools[0].name == "echo_tool"
        assert tools[0].invoke("hello") == "Echo: hello"


def test_tool_name_with_keyword():
    mcp_server_script = dedent(
        '''
        from mcp.server.fastmcp import FastMCP

        mcp = FastMCP("Echo Server")

        @mcp.tool(name="def")
        def echo_tool(text: str) -> str:
            """Echo the input text"""
            return f"Echo: {text}"
        
        mcp.run()
        '''
    )
    with MCPAdapt(
        StdioServerParameters(
            command="uv", args=["run", "python", "-c", mcp_server_script]
        ),
        LangChainAdapter(),
    ) as tools:
        assert len(tools) == 1
        assert tools[0].name == "def_"
        assert tools[0].invoke("hello") == "Echo: hello"



"""This module implements the LangChain adapter.

LangChain tools support both sync and async functions for their tools so we can
leverage both in our implementation.

Example Usage:
>>> with MCPAdapt(StdioServerParameters(command="uv", args=["run", "src/echo.py"]), LangChainAdapter()) as tools:
>>>     print(tools)
"""

import keyword
import logging
import re
from functools import partial
from typing import Any, Callable, Coroutine

import jsonref
import langchain_core
import mcp
from langchain.tools import BaseTool

from mcpadapt.core import ToolAdapter

log = logging.getLogger(__name__)

JSON_SCHEMA_TO_PYTHON_TYPES = {
    "string": "str",
    "number": "float",
    "integer": "int",
    "object": "dict",
    "array": "list",
    "boolean": "bool",
    "null": "None",
}


def _sanitize_function_name(name):
    """
    A function to sanitize function names to be used as a tool name.
    Prevent the use of dashes or other python keywords as function names by tool.
    """
    # Replace dashes with underscores
    name = name.replace("-", "_")

    # Remove any characters that aren't alphanumeric or underscore
    name = re.sub(r"[^\w_]", "", name)

    # Ensure it doesn't start with a number
    if name[0].isdigit():
        name = f"_{name}"

    # Check if it's a Python keyword
    if keyword.iskeyword(name):
        name = f"{name}_"

    return name


def _generate_tool_class(
    name: str,
    description: str,
    input_schema: dict[str, Any],
    async_func: bool = False,
    parse_docstring: bool = True,
) -> str:
    """Generate a tool BaseTool class for `langchain` from MCP tool information.

    Note we use the simpliest '@tool' decorator for now.

    Args:
        name: the name of the tool as used in the MCP protocol
        description: the description of the tool as used in the MCP protocol
        input_schema: the input schema of the tool as used in the MCP protocol
        async_func: whether the function is async or not
        parse_docstring: whether to parse the docstring as a Google-Style docstring

    Returns:
        the generated langchain tool class as a string to be executed with exec.
    """
    resolved_json_schema = jsonref.replace_refs(input_schema)
    properties = resolved_json_schema.get("properties", {})

    # construct typed signature based on input schema
    # TODO: this could be better and handle nested objects...
    tool_params = []
    for k, v in properties.items():
        tool_params.append(f"{k}: {JSON_SCHEMA_TO_PYTHON_TYPES[v['type']]}")
    tool_params = ", ".join(tool_params)

    argument = "{" + ", ".join(f"'{k}': {k}" for k in properties.keys()) + "}"

    # change def statement and return statement based on async_func
    def_statement = "def"
    return_statement = f"return func({argument}).content[0].text"
    if async_func:
        def_statement = "async def"
        return_statement = f"return (await func({argument})).content[0].text"

    # define the decorator based on parse_docstring
    decorator = "@tool(parse_docstring=True)" if parse_docstring else "@tool"

    class_template = f'''
{decorator}
{def_statement} {name}({tool_params}) -> str:
    """{description}"""
    {return_statement}
'''.strip()

    return class_template


def _instanciate_tool(
    mcp_tool_name: str,
    generate_class_template: Callable[[bool], str],
    func: Callable[[dict | None], mcp.types.CallToolResult]
    | Callable[[dict | None], Coroutine[Any, Any, mcp.types.CallToolResult]],
) -> BaseTool:
    """Instanciate a tool from a class template and a function wrapping the mcp tool_call.

    Args:
        mcp_tool_name: the name of the tool as used in the MCP protocol
        generate_class_template: a function that generates the class template
            (with or without parsing the docstring)
        func: the function wrapping the mcp tool_call

    Returns:
        the instanciated langchain tool
    """

    # Create namespace and execute the class definition
    namespace = {"tool": langchain_core.tools.tool, "func": func}
    try:
        exec(generate_class_template(True), namespace)
    except ValueError as e:
        if "Found invalid Google-Style docstring." in str(e):
            exec(generate_class_template(False), namespace)

    # Get the class from namespace and instantiate it
    tool = namespace[mcp_tool_name]
    return tool


class LangChainAdapter(ToolAdapter):
    """Adapter for `langchain`.

    Note that `langchain` support both sync and async tools so we
    write adapt for both methods.

    Warning: if the mcp tool name is a python keyword, starts with digits or contains
    dashes, the tool name will be sanitized to become a valid python function name.

    """

    def adapt(
        self,
        func: Callable[[dict | None], mcp.types.CallToolResult],
        mcp_tool: mcp.types.Tool,
    ) -> BaseTool:
        mcp_tool_name = _sanitize_function_name(mcp_tool.name)
        if mcp_tool_name != mcp_tool.name:
            log.warning(f"MCP tool name {mcp_tool.name} sanitized to {mcp_tool_name}")

        generate_class_template = partial(
            _generate_tool_class,
            mcp_tool_name,
            mcp_tool.description,
            mcp_tool.inputSchema,
            False,
        )
        return _instanciate_tool(mcp_tool_name, generate_class_template, func)

    def async_adapt(
        self,
        afunc: Callable[[dict | None], Coroutine[Any, Any, mcp.types.CallToolResult]],
        mcp_tool: mcp.types.Tool,
    ) -> BaseTool:
        mcp_tool_name = _sanitize_function_name(mcp_tool.name)
        if mcp_tool_name != mcp_tool.name:
            log.warning(f"MCP tool name {mcp_tool.name} sanitized to {mcp_tool_name}")

        generate_class_template = partial(
            _generate_tool_class,
            mcp_tool_name,
            mcp_tool.description,
            mcp_tool.inputSchema,
            True,
        )
        return _instanciate_tool(mcp_tool_name, generate_class_template, afunc)


if __name__ == "__main__":
    import asyncio

    from mcp import StdioServerParameters

    from mcpadapt.core import MCPAdapt

    with MCPAdapt(
        StdioServerParameters(command="uv", args=["run", "src/echo.py"]),
        LangChainAdapter(),
    ) as tools:
        print(tools)
        print(tools[0].invoke("hello"))

    async def main():
        async with MCPAdapt(
            StdioServerParameters(command="uv", args=["run", "src/echo.py"]),
            LangChainAdapter(),
        ) as tools:
            print(tools)
            print(await tools[0].ainvoke("hello"))

    asyncio.run(main())



"""An example of using the LangChain adapter to adapt MCP tools to LangChain tools.

This example uses the PubMed API to search for studies.
"""

import os

# Import relevant functionality
from langchain_anthropic import ChatAnthropic
from langchain_core.messages import HumanMessage
from langgraph.checkpoint.memory import MemorySaver
from langgraph.prebuilt import create_react_agent
from mcp import StdioServerParameters

from mcpadapt.core import MCPAdapt
from mcpadapt.langchain_adapter import LangChainAdapter
from dotenv import load_dotenv

load_dotenv()
if not os.environ.get("ANTHROPIC_API_KEY"):
    raise ValueError(
        "ANTHROPIC_API_KEY is not set create a .env file at the root of the project with ANTHROPIC_API_KEY=<your-api-key>"
    )


def main():
    """Fully synchronous version example. Note that async is preferred (just below).
    As it doesnt rely on the hack with a separate thread running the MCP server.
    """
    with MCPAdapt(
        StdioServerParameters(
            command="uvx",
            args=["--quiet", "pubmedmcp@0.1.3"],
            env={"UV_PYTHON": "3.12", **os.environ},
        ),
        LangChainAdapter(),
    ) as tools:
        # Create the agent
        memory = MemorySaver()

        model = ChatAnthropic(
            model_name="claude-3-5-sonnet-20241022", max_tokens_to_sample=8192
        )
        agent_executor = create_react_agent(model, tools, checkpointer=memory)

        # Use the agent
        config = {"configurable": {"thread_id": "abc123"}}
        for chunk in agent_executor.stream(
            {
                "messages": [
                    HumanMessage(
                        content="Find relevant studies on alcohol hangover and treatment."
                    )
                ]
            },
            config,
        ):
            print(chunk)
            print("----")

        for chunk in agent_executor.stream(
            {"messages": [HumanMessage(content="whats the weather where I live?")]},
            config,
        ):
            print(chunk)
            print("----")


async def async_main():
    """Fully asynchronous version example."""
    async with MCPAdapt(
        StdioServerParameters(
            command="uvx",
            args=["--quiet", "pubmedmcp@0.1.3"],
            env={"UV_PYTHON": "3.12", **os.environ},
        ),
        LangChainAdapter(),
    ) as tools:
        # Create the agent
        memory = MemorySaver()

        model = ChatAnthropic(
            model_name="claude-3-5-sonnet-20241022", max_tokens_to_sample=8192
        )
        agent_executor = create_react_agent(model, tools, checkpointer=memory)

        # Use the agent
        config = {"configurable": {"thread_id": "abc123"}}
        async for event in agent_executor.astream(
            {
                "messages": [
                    HumanMessage(
                        content="Find relevant studies on alcohol hangover and treatment."
                    )
                ]
            },
            config,
        ):
            print(event)
            print("----")

        async for event in agent_executor.astream(
            {"messages": [HumanMessage(content="whats the weather where I live?")]},
            config,
        ):
            print(event)
            print("----")


if __name__ == "__main__":
    import asyncio

    # run the sync version
    # main()

    # run the async version
    asyncio.run(async_main())



#!/usr/bin/env python
import re
import sys
import subprocess
from pathlib import Path


def run_command(command):
    try:
        subprocess.run(command, check=True, shell=True)
    except subprocess.CalledProcessError as e:
        print(f"Error executing command: {command}")
        print(f"Error: {e}")
        sys.exit(1)


def bump_version(version_type):
    init_file = Path("src/mcpadapt/__init__.py")

    # Read current version
    content = init_file.read_text()
    current_version = re.search(r'__version__ = ["\']([^"\']+)["\']', content).group(1)
    major, minor, patch = map(int, current_version.split("."))

    # Update version based on argument
    if version_type == "major":
        new_version = f"{major + 1}.0.0"
    elif version_type == "minor":
        new_version = f"{major}.{minor + 1}.0"
    elif version_type == "patch":
        new_version = f"{major}.{minor}.{patch + 1}"
    else:
        print("Invalid version type. Use 'major', 'minor', or 'patch'")
        sys.exit(1)

    # Update __init__.py
    new_content = re.sub(
        r'__version__ = ["\']([^"\']+)["\']', f'__version__ = "{new_version}"', content
    )
    init_file.write_text(new_content)

    # Git operations
    run_command("git add src/mcpadapt/__init__.py")
    run_command(f'git commit -m "release {new_version}: version bump commit"')
    run_command("git push")
    run_command(f"git tag v{new_version}")
    run_command("git push --tags")

    # Create GitHub release using gh CLI
    run_command(
        f'gh release create v{new_version} --title "Release {new_version}" --generate-notes'
    )

    print(f"Version bumped from {current_version} to {new_version}")
    print(f"Git operations completed and GitHub release v{new_version} created")


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print("Usage: bump_version.py <major|minor|patch>")
        sys.exit(1)

    bump_version(sys.argv[1])



"""Core module for the MCPAdapt library.

This module contains the core functionality for the MCPAdapt library. It provides the
basic interfaces and classes for adapting tools from MCP to the desired Agent framework.
"""

import asyncio
import threading
from abc import ABC, abstractmethod
from contextlib import asynccontextmanager
from functools import partial
from typing import Any, Callable, Coroutine

import mcp
from mcp import ClientSession, StdioServerParameters
from mcp.client.stdio import stdio_client


class ToolAdapter(ABC):
    """A basic interface for adapting tools from MCP to the desired Agent framework."""

    @abstractmethod
    def adapt(
        self,
        func: Callable[[dict | None], mcp.types.CallToolResult],
        mcp_tool: mcp.types.Tool,
    ):
        """Adapt a single tool from MCP to the desired Agent framework.

        The MCP protocol will provide a name, description and inputSchema in JSON Schema
        format. This needs to be adapted to the desired Agent framework.

        Note that the function is synchronous (not a coroutine) you can use
        :meth:`ToolAdapter.async_adapt` if you need to use the tool asynchronously.

        Args:
            func: The function to be called.
            mcp_tool: The tool to adapt.

        Returns:
            The adapted tool.
        """
        pass

    def async_adapt(
        self,
        afunc: Callable[[dict | None], Coroutine[Any, Any, mcp.types.CallToolResult]],
        mcp_tool: mcp.types.Tool,
    ):
        """Adapt a single tool from MCP to the desired Agent framework.

        The MCP protocol will provide a name, description and inputSchema in JSON Schema
        format. This needs to be adapted to the desired Agent framework.

        Note that the function is asynchronous (a coroutine) you can use
        :meth:`ToolAdapter.adapt` if you need to use the tool synchronously.

        Args:
            afunc: The coroutine to be called.
            mcp_tool: The tool to adapt.

        Returns:
            The adapted tool.
        """
        raise NotImplementedError(
            "Async adaptation is not supported for this Agent framework."
        )


@asynccontextmanager
async def mcptools(
    serverparams: StdioServerParameters,
) -> tuple[ClientSession, list[mcp.types.Tool]]:
    """Async context manager that yields tools from an MCP server.

    Note: the session can be then used to call tools on the MCP server but it's async.
    Use MCPAdapt instead if you need to use the tools synchronously.

    Args:
        serverparams: Parameters to run the MCP server.

    Yields:
        A list of tools available on the MCP server.

    Usage:
    >>> async with mcptools(StdioServerParameters(command="uv", args=["run", "src/echo.py"])) as (session, tools):
    >>>     print(tools)
    """
    async with stdio_client(serverparams) as (read, write):
        async with ClientSession(read, write) as session:
            # Initialize the connection and get the tools from the mcp server
            await session.initialize()
            tools = await session.list_tools()
            yield session, tools.tools


class MCPAdapt:
    """The main class for adapting MCP tools to the desired Agent framework.

    This class can be used either as a sync or async context manager.

    If running synchronously, it will run the MCP server in a separate thread and take
    care of making the tools synchronous without blocking the server.

    If running asynchronously, it will use the async context manager and return async
    tools.

    Dependening on what your Agent framework supports choose the approriate method. If
    async is supported it is recommended.

    Important Note: adapters need to implement the async_adapt method to support async
    tools.

    Usage:
    >>> # sync usage
    >>> with MCPAdapt(StdioServerParameters(command="uv", args=["run", "src/echo.py"]), SmolAgentAdapter()) as tools:
    >>>     print(tools)

    >>> # async usage
    >>> async with MCPAdapt(StdioServerParameters(command="uv", args=["run", "src/echo.py"]), SmolAgentAdapter()) as tools:
    >>>     print(tools)
    """

    def __init__(self, serverparams: StdioServerParameters, adapter: ToolAdapter):
        # attributes we receive from the user.
        self.serverparams = serverparams
        self.adapter = adapter

        # session and tools get set by the async loop during initialization.
        self.session: ClientSession = None
        self.mcp_tools: list[mcp.types.Tool] = None

        # all attributes used to manage the async loop and separate thread.
        self.loop = asyncio.new_event_loop()
        self.task = None
        self.ready = threading.Event()
        self.thread = threading.Thread(target=self._run_loop, daemon=True)

        # start the loop in a separate thread and wait till ready synchronously.
        self.thread.start()
        self.ready.wait()

    def _run_loop(self):
        """Runs the event loop in a separate thread (for synchronous usage)."""
        asyncio.set_event_loop(self.loop)

        async def setup():
            async with mcptools(self.serverparams) as (session, tools):
                self.session, self.mcp_tools = session, tools
                self.ready.set()  # Signal initialization is complete
                await asyncio.Event().wait()  # Keep session alive until stopped

        self.task = self.loop.create_task(setup())
        try:
            self.loop.run_until_complete(self.task)
        except asyncio.CancelledError:
            pass

    def tools(self) -> list[Any]:
        """Returns the tools from the MCP server adapted to the desired Agent framework.

        This is what is yielded if used as a context manager otherwise you can access it
        directly via this method.

        An equivalent async method is available if your Agent framework supports it:
        see :meth:`atools`.

        """
        if not self.session:
            raise RuntimeError("Session not initialized")

        def _sync_call_tool(
            name: str, arguments: dict | None = None
        ) -> mcp.types.CallToolResult:
            return asyncio.run_coroutine_threadsafe(
                self.session.call_tool(name, arguments), self.loop
            ).result()

        return [
            self.adapter.adapt(partial(_sync_call_tool, tool.name), tool)
            for tool in self.mcp_tools
        ]

    def close(self):
        """Clean up resources and stop the client."""
        if self.task and not self.task.done():
            self.loop.call_soon_threadsafe(self.task.cancel)
        self.thread.join()  # will wait until the task is cancelled to join thread (as it's blocked Event().wait())
        self.loop.close()  # we won't be using the loop anymore we can safely close it

    def __enter__(self):
        return self.tools()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    # -- add support for async context manager as well if the agent framework supports it.
    def atools(self) -> list[Any]:
        """Returns the tools from the MCP server adapted to the desired Agent framework.

        This is what is yielded if used as an async context manager otherwise you can
        access it directly via this method.

        An equivalent async method is available if your Agent framework supports it:
        see :meth:`atools`.
        """
        return [
            self.adapter.async_adapt(partial(self.session.call_tool, tool.name), tool)
            for tool in self.mcp_tools
        ]

    async def __aenter__(self) -> list[Any]:
        self._ctxmanager = mcptools(self.serverparams)
        self.session, self.tools = await self._ctxmanager.__aenter__()
        return self.atools()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._ctxmanager.__aexit__(exc_type, exc_val, exc_tb)


if __name__ == "__main__":

    class DummyAdapter(ToolAdapter):
        def adapt(
            self,
            func: Callable[[dict | None], mcp.types.CallToolResult],
            mcp_tool: mcp.types.Tool,
        ):
            return func

        def async_adapt(
            self,
            afunc: Callable[
                [dict | None], Coroutine[Any, Any, mcp.types.CallToolResult]
            ],
            mcp_tool: mcp.types.Tool,
        ):
            return afunc

    with MCPAdapt(
        StdioServerParameters(command="uv", args=["run", "src/echo.py"]),
        DummyAdapter(),
    ) as smolagents_tools:
        print(smolagents_tools)
        print(smolagents_tools[0].forward({"text": "hello"}))

    async def main():
        async with MCPAdapt(
            StdioServerParameters(command="uv", args=["run", "src/echo.py"]),
            DummyAdapter(),
        ) as smolagents_tools:
            print(smolagents_tools)
            print(smolagents_tools[0].forward({"text": "hello"}))

    asyncio.run(main())



from textwrap import dedent

import pytest
from mcp import StdioServerParameters

from mcpadapt.core import MCPAdapt
from mcpadapt.smolagents_adapter import SmolAgentsAdapter


@pytest.fixture
def echo_server_script():
    return dedent(
        '''
        from mcp.server.fastmcp import FastMCP

        mcp = FastMCP("Echo Server")

        @mcp.tool()
        def echo_tool(text: str) -> str:
            """Echo the input text"""
            return f"Echo: {text}"
        
        mcp.run()
        '''
    )


def test_basic_sync(echo_server_script):
    with MCPAdapt(
        StdioServerParameters(
            command="uv", args=["run", "python", "-c", echo_server_script]
        ),
        SmolAgentsAdapter(),
    ) as tools:
        assert len(tools) == 1
        assert tools[0].name == "echo_tool"
        assert tools[0]("hello") == "Echo: hello"



"""An example of using MCPAdapt with the smolagents framework to query pubmed papers.

Note this is just a demo MCP server that I implemented for the purpose of this example.
It only provide a single tool to search amongst pubmed papers abstracts.

Note that this example needs the smolagents optional dependency to run.

Usage:
>>> uv add mcpadapt[smolagents]
>>> uv run examples/smolagents_pubmed.py
"""

import os

from mcp import StdioServerParameters
from smolagents import CodeAgent, HfApiModel

from mcpadapt.core import MCPAdapt
from mcpadapt.smolagents_adapter import SmolAgentsAdapter

with MCPAdapt(
    StdioServerParameters(
        command="uvx",
        args=["--quiet", "pubmedmcp@0.1.3"],
        env={"UV_PYTHON": "3.12", **os.environ},
    ),
    SmolAgentsAdapter(),
) as tools:
    print(tools[0](request={"term": "efficient treatment hangover"}))
    agent = CodeAgent(tools=tools, model=HfApiModel())
    agent.run("Find studies about hangover?")



"""
FastMCP dummy Echo Server for testing purposes.
"""

from mcp.server.fastmcp import FastMCP

# Create server
mcp = FastMCP("Echo Server")


@mcp.tool()
def echo_tool(text: str) -> str:
    """Echo the input text

    Args:
        text (str): The text to echo

    Returns:
        str: The echoed text
    """
    return text


@mcp.resource("echo://static")
def echo_resource() -> str:
    return "Echo!"


@mcp.resource("echo://{text}")
def echo_template(text: str) -> str:
    """Echo the input text"""
    return f"Echo: {text}"


@mcp.prompt("echo")
def echo_prompt(text: str) -> str:
    return text


if __name__ == "__main__":
    import sys

    print("starting echo server", file=sys.stderr)
    mcp.run()
    print("echo server ended", file=sys.stderr)



"""This module implements the SmolAgents adapter.

SmolAgents do not support async tools, so this adapter will only work with the sync
context manager.

Example Usage:
>>> with MCPAdapt(StdioServerParameters(command="uv", args=["run", "src/echo.py"]), SmolAgentAdapter()) as tools:
>>>     print(tools)
"""

import json
from typing import Any, Callable

import jsonref
import mcp
import smolagents

from mcpadapt.core import ToolAdapter


def _generate_tool_inputs(resolved_json_schema: dict[str, Any]) -> dict[str, str]:
    """
    takes an json_schema as used in the MCP protocol and return an inputs dict for
    smolagents tools. see AUTHORIZED_TYPES in smolagents.tools for the types allowed.
    Note that we consider the json_schema to already have $ref resolved with jsonref for
    example.
    """
    return {
        # TODO: use google-docstring-parser to parse description of args and pass it here...
        k: {"type": v["type"], "description": v.get("description", "")}
        for k, v in resolved_json_schema.items()
    }


def _generate_tool_class(
    name: str,
    description: str,
    inputSchema: dict[str, Any],
) -> str:
    """generate a smolagents tool class from the MCP protocol informations.

    Note: we generate code because smolagents is very picky about the signature of the
    forward method matching the inputs name and type.

    Return a string with the class so it's easy to test and debug the generated code.

    Args:
        name: the name of the tool as used in the MCP protocol
        description: the description of the tool as used in the MCP protocol
        inputSchema: the input schema of the tool as used in the MCP protocol

    Returns:
        the generated smolagentstool class as a string to be executed with exec.
        **important**: the class_template need smolagents and func in the namespace to
        be exec. We assume func as being a sync function taking a single dict argument
        and returning a CallToolResult as in:
        func: Callable[[dict | None], mcp.types.CallToolResult]
    """
    resolved_json_schema = jsonref.replace_refs(inputSchema).get("properties", {})
    smolagents_inputs = _generate_tool_inputs(resolved_json_schema)

    # smolagents provide arguments to the forward as follow forward(arg1=..., arg2=...)
    # but MCP call_tool takes a single 'argument' as in func({'arg1': .., 'arg2': ..})
    forward_params = ", ".join(f"{k}" for k in smolagents_inputs.keys())
    argument = "{" + ", ".join(f"'{k}': {k}" for k in smolagents_inputs.keys()) + "}"

    class_template = f'''
class SmolAgentsTool(smolagents.Tool):
    def __init__(self):
        super().__init__()
        self.name = "{name}"
        self.description = """{description}"""
        self.inputs = {json.dumps(smolagents_inputs)}
        self.output_type = "string"
    
    def forward(self, {forward_params}) -> str:
        """
        Forward method with dynamically generated parameters and argument.
        """
        return func({argument}).content[0].text
'''.strip()

    return class_template


class SmolAgentsAdapter(ToolAdapter):
    """Adapter for the `smolagents` framework.

    Note that the `smolagents` framework do not support async tools at this time so we
    write only the adapt method.
    """

    def adapt(
        self,
        func: Callable[[dict | None], mcp.types.CallToolResult],
        mcp_tool: mcp.types.Tool,
    ) -> smolagents.Tool:
        class_template = _generate_tool_class(
            mcp_tool.name, mcp_tool.description, mcp_tool.inputSchema
        )

        # Create namespace and execute the class definition
        namespace = {"smolagents": smolagents, "func": func}
        exec(class_template, namespace)

        # Get the class from namespace and instantiate it
        tool = namespace["SmolAgentsTool"]()
        return tool


if __name__ == "__main__":
    from mcp import StdioServerParameters

    from mcpadapt.core import MCPAdapt

    with MCPAdapt(
        StdioServerParameters(command="uv", args=["run", "src/echo.py"]),
        SmolAgentsAdapter(),
    ) as tools:
        print(tools)
